/*
 * serve-index
 * Copyright(c) 2011 Sencha Inc.
 * Copyright(c) 2011 TJ Holowaychuk
 * Copyright(c) 2014-2015 Douglas Christopher Wilson
 * https://github.com/expressjs/serve-index
 *
 * serve-directory
 * Copyright(c) 2017- fisker Cheung
 * MIT Licensed
 */

import { normalize, sep, join, extname, resolve } from 'path';
import { statSync, readdirSync, readFileSync, existsSync } from 'fs';
import httpError from 'http-errors';
import accepts from 'accepts';
import { original } from 'parseurl';
import debug$1 from 'debug';
import { lookup } from 'mime-types';
import _template from 'lodash.template';

function _defineProperty(obj, key, value) {
  if (key in obj) {
    Object.defineProperty(obj, key, {
      value: value,
      enumerable: true,
      configurable: true,
      writable: true
    });
  } else {
    obj[key] = value;
  }

  return obj;
}

function ownKeys(object, enumerableOnly) {
  var keys = Object.keys(object);

  if (Object.getOwnPropertySymbols) {
    var symbols = Object.getOwnPropertySymbols(object);
    if (enumerableOnly) symbols = symbols.filter(function (sym) {
      return Object.getOwnPropertyDescriptor(object, sym).enumerable;
    });
    keys.push.apply(keys, symbols);
  }

  return keys;
}

function _objectSpread2(target) {
  for (var i = 1; i < arguments.length; i++) {
    var source = arguments[i] != null ? arguments[i] : {};

    if (i % 2) {
      ownKeys(source, true).forEach(function (key) {
        _defineProperty(target, key, source[key]);
      });
    } else if (Object.getOwnPropertyDescriptors) {
      Object.defineProperties(target, Object.getOwnPropertyDescriptors(source));
    } else {
      ownKeys(source).forEach(function (key) {
        Object.defineProperty(target, key, Object.getOwnPropertyDescriptor(source, key));
      });
    }
  }

  return target;
}

var name = "serve-directory";
var version = "3.1.1";

var debug = debug$1(name);

function mime(extension) {
  return lookup(extension) || '';
}

function directoryFirst(a, b) {
  return b.isDirectory() - a.isDirectory();
}

function sortBy(key) {
  return function (a, b) {
    if (b[key] === a[key]) {
      return 0;
    }

    return b[key] < a[key] ? 1 : -1;
  };
}

function sortFile(a, b) {
  return directoryFirst(a, b) || sortBy('name')(a, b);
}

function startsWithDot(file) {
  return file[0] === '.';
}

function isHiddenFile(file) {
  return startsWithDot(file.name);
}

function notHiddenFile(file) {
  return !isHiddenFile(file);
}

function isHiddenPath(path) {
  return normalize(path).split(sep).some(startsWithDot);
}

class Connection {
  constructor(sd, request, response, next) {
    this.sd = sd;
    this.request = request;
    this.response = response;
    this.next = next;
    this.url = original(this.request);
  }

  getMethod() {
    const method = this.request.method;

    if (method !== 'GET' && method !== 'HEAD') {
      this.response.statusCode = method === 'OPTIONS' ? 200 : 405;
      this.response.setHeader('Allow', 'GET, HEAD, OPTIONS');
      this.response.setHeader('Content-Length', '0');
      this.response.end();
      return null;
    }

    this.method = method;
    return method;
  }

  getPathname() {
    let pathname = this.url.pathname;
    const hidden = this.sd.options.hidden;

    try {
      pathname = decodeURIComponent(pathname);
    } catch (error) {
      this.next(httpError(400));
    }

    if (!hidden && isHiddenPath(pathname)) {
      debug('hidden folder "%s" deny.', pathname);
      this.next(httpError(403));
      return null;
    } // null byte(s), bad request


    if (pathname.includes('\0')) {
      debug('null byte(s) in "%s", bad request.', pathname);
      this.next(httpError(400));
      return null;
    }

    pathname = pathname.replace(/\/+/g, '/');
    this.pathname = pathname;
    return pathname;
  }

  getResponseType() {
    const acceptMediaTypes = Object.keys(this.sd.responser);
    const responseType = accepts(this.request).type(acceptMediaTypes);

    if (!responseType) {
      debug('mime not acceptable "%s".', responseType);
      this.next(httpError(406));
      return null;
    }

    this.responseType = responseType;
    return responseType;
  }

  getResponser() {
    const responser = this.sd.responser[this.responseType];
    return this.responser = responser;
  }

  getPath() {
    // join / normalize from root dir
    const path = join(this.sd.root, this.pathname); // malicious path

    if (!path.startsWith(this.sd.root + sep)) {
      debug('malicious path "%s".', this.pathname);
      this.next(httpError(403));
      return null;
    }

    this.path = path;
    return path;
  }

  getDirectory() {
    debug('get directory "%s".', this.path);
    let stats;

    try {
      stats = statSync(this.path);
    } catch (error) {
      if (error.code === 'ENOENT' || error.code === 'ENOTDIR') {
        this.next();
        return null;
      }

      error.status = error.code === 'ENAMETOOLONG' ? 414 : 500;
      this.next(error);
      return null;
    }

    if (!stats.isDirectory()) {
      this.next();
      return null;
    }

    if (this.pathname.slice(-1) !== '/') {
      debug('add "/" to "%s".', this.pathname);
      this.response.writeHead(301, {
        Location: `${this.url.pathname}/`
      });
      this.response.end();
      return null;
    }

    stats.path = this.path;
    stats.pathname = this.pathname;
    stats.url = this.sd.options.relative ? '.' : this.url.pathname;
    this.directory = stats;
    return stats;
  }

  getFiles() {
    debug('get files "%s"', this.path);
    const path = this.path;
    const urlPrefix = this.sd.options.relative ? '' : this.url.pathname;
    let files;

    try {
      files = readdirSync(path).map(function (file) {
        const stats = statSync(join(path, file));
        stats.name = file;
        stats.ext = extname(file);
        stats.type = mime(stats.ext);
        stats.url = urlPrefix + encodeURIComponent(file) + (stats.isDirectory() ? '/' : '');
        return stats;
      }).sort(sortFile);
    } catch (error) {
      this.next(error);
      return null;
    }

    if (!this.sd.options.hidden) {
      files = files.filter(notHiddenFile);
    }

    this.files = files;
    return files;
  }

  process() {
    if (!this.getMethod() || !this.getPathname() || !this.getResponseType() || !this.getResponser() || !this.getPath() || !this.getDirectory() || !this.getFiles()) {
      return;
    }

    try {
      this.responser(this.request, this.response, {
        path: this.path,
        pathname: this.pathname,
        url: this.url,
        method: this.method,
        responseType: this.responseType,
        directory: this.directory,
        files: this.files,
        fileNames: this.files.map(({
          name
        }) => name)
      });
    } catch (error) {
      error.status = 500;
      this.next(error);
    }
  }

}

const CHARSET = 'utf-8';
const EOL = '\n';
const DEFAULT_HTML_TEMPLATE_FILE = join(__dirname, 'directory.jst');

var defaults = {
  hidden: false,
  relative: true,
  process: [{
    accept: 'text/html',
    render: DEFAULT_HTML_TEMPLATE_FILE
  }, {
    accept: 'text/plain',
    render: ({
      fileNames
    }) => fileNames.join(EOL) + EOL
  }, {
    accept: 'application/json',
    render: ({
      fileNames
    }) => JSON.stringify(fileNames)
  }]
};

function identity(x) {
  return x;
}

function isBoolean(x) {
  return x === true || x === false;
}

function isFunction(x) {
  return typeof x === 'function';
}

function isString(x) {
  return typeof x === 'string';
}

function noop() {}

function readFile(file) {
  try {
    return readFileSync(file, CHARSET);
  } catch (_unused) {}

  return null;
}

const toString = Object.prototype.toString;

function type(x) {
  return toString.call(x).slice(8, -1);
}



var utils = /*#__PURE__*/Object.freeze({
  directoryFirst: directoryFirst,
  identity: identity,
  isHiddenFile: isHiddenFile,
  isHiddenPath: isHiddenPath,
  isBoolean: isBoolean,
  isFunction: isFunction,
  isString: isString,
  mime: mime,
  noop: noop,
  notHiddenFile: notHiddenFile,
  readFile: readFile,
  sortBy: sortBy,
  sortFile: sortFile,
  type: type
});

const package_ = {
  name,
  version
};
var imports = Object.freeze(_objectSpread2({
  CHARSET,
  pkg: package_,
  require
}, utils));

function template(template, options) {
  if (isFunction(template)) {
    return template;
  }

  if (isString(template)) {
    if (existsSync(template)) {
      template = readFile(template);
    }

    return _template(template, options || {
      imports
    });
  }

  return identity;
}

function responser(mime, render) {
  return (request, response, data) => {
    const body = render(data); // security header for content sniffing

    response.setHeader('X-Content-Type-Options', 'nosniff'); // standard headers

    response.setHeader('Content-Type', `${mime}; charset=${CHARSET}`);
    response.setHeader('Content-Length', Buffer.byteLength(body, CHARSET)); // body

    response.end(body, CHARSET);
  };
}

class ServeDirectory {
  constructor(root, options) {
    // root required
    // resolve root to absolute and normalize
    root = resolve(root);
    this.root = root;
    this.responser = {};
    this.options = {
      hidden: defaults.hidden,
      relative: defaults.relative
    };
    this.imports = _objectSpread2({}, imports);
    this.config(defaults);

    if (options) {
      this.config(options);
    }
  }

  config({
    hidden,
    relative,
    imports = {},
    process = []
  } = {}) {
    if (isBoolean(hidden)) {
      this.options.hidden = hidden;
    }

    if (isBoolean(relative)) {
      this.options.relative = relative;
    }

    Object.assign(this.imports, imports);
    process.filter(Boolean).forEach(({
      accept,
      render
    }) => {
      accept = (Array.isArray(accept) ? accept : accept.split(',')).map(x => x.trim()).filter(Boolean);
      var _iteratorNormalCompletion = true;
      var _didIteratorError = false;
      var _iteratorError = undefined;

      try {
        for (var _iterator = accept[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
          const type = _step.value;

          if (render) {
            this.responser[type] = responser(type, template(render, {
              imports: this.imports
            }));
          } else {
            delete this.responser[type];
          }
        }
      } catch (err) {
        _didIteratorError = true;
        _iteratorError = err;
      } finally {
        try {
          if (!_iteratorNormalCompletion && _iterator.return != null) {
            _iterator.return();
          }
        } finally {
          if (_didIteratorError) {
            throw _iteratorError;
          }
        }
      }
    });
  }

  middleware(request, response, next) {
    try {
      return new Connection(this, request, response, next).process();
    } catch (error) {
      return next(error);
    }
  }

}

function middleware(root, options) {
  const sd = new ServeDirectory(root, options);
  return sd.middleware.bind(sd);
}

export default middleware;
